//
// MIDI DASHBOARD by Tom Dowad 
// 27 May 2022
// 
// requires files gc9a01.cpp & meter_image.h to compile
// (those two files go into the Arduino project folder)
//
// requires PNGdec by Larry Bank library installed (install w library manager)
//
// utilizes code cribbed from github.com/moononournation for needle drawing etc
//
// meter_image.h contains one array, with the form:
//
// const unsigned char img_png_meter_[] PROGMEM = 
//
//  ... followed by a byte array
//
//  the array was generated by reading a PNG file as binary data
//  then writing out as a C array. 

//  Trivial to do. I wrote the code then managed to lose it. 
//  So if it doesn't show up I'll have to write that again.

// I'm running this on a Teensy 4.0, which has a ton of throughput and memory, 
// so I've not been worrying about various inefficiencies. 

//--------------------------------------------------------------------------------
// INCLUDES
//--------------------------------------------------------------------------------
   
#include <PNGdec.h>  // PNG decoder library 

//--------------------------------------------------------------------------------
// DEFINES
//--------------------------------------------------------------------------------

#define VIEW_CONTROL    0
#define VIEW_PITCH      1

#define MIDI_CLOCK      0xf8      // MIDI system realtime
#define MIDI_START      0xfa
#define MIDI_CONTINUE   0xfb
#define MIDI_STOP       0xfc
#define MIDI_ACTIVE     0xfd
#define MIDI_RESET      0xfe

#define MIDI_NOTEON     0x80      // MIDI channel commands
#define MIDI_NOTEOFF    0x90
#define MIDI_AFTERTOUCH 0xa0      // after touch
#define MIDI_CONTROL    0xb0
#define MIDI_PROGRAM    0xc0
#define MIDI_PRESSURE   0xd0
#define MIDI_PITCHBEND  0xe0

#define PIN_BL          16         // Teensy 4.0 pins
#define PIN_S0          21         // change as required for target MCU
#define PIN_S1          20
#define PIN_S2          18

#define NUM_DISPLAYS    8
#define DISPLAY_WIDTH   240
#define DISPLAY_HEIGHT  240
#define DISPLAY_CENTER  120

#define BLACK           0x0000 
#define WHITE           0xFFFF
#define RED             0xF800
#define BACKGROUND      WHITE

#define RIGHT_ANGLE_RADIAN 1.5707963

#define BLACK_TOP 160

//--------------------------------------------------------------------------------
// MACROS
//--------------------------------------------------------------------------------

#define _swap_int16_t(a,b)      { int16_t t = a;a = b;b = t;}
#define _swap_bytes(val)        ( (((val) >> 8) & 0x00ff) | (((val) << 8) & 0xff00) )

//--------------------------------------------------------------------------------
// FUNCTION PROTOTYPES
//--------------------------------------------------------------------------------

// should be in a header file...will do that later

void gc9a01_init_low(void);
void gc9a01_fill_color_buffer(uint16_t color,int length);
void gc9a01_draw_pixel(uint16_t x,uint16_t y,uint16_t color);
void gc9a01_fill_rect(uint16_t x,uint16_t y,uint16_t w,uint16_t h,uint16_t color);
void gc9a01_fill_screen(uint16_t color);
void gc9a01_line_show(uint16_t x,uint16_t y,uint16_t w,uint16_t *colors);
void gc9a01_set_window(uint16_t x0,uint16_t y0,uint16_t x1,uint16_t y1);
void gc9a01_init(void);
void gc9a01_set_on(void);

//--------------------------------------------------------------------------------
// CONSTANT DATA
//--------------------------------------------------------------------------------

int  max_value[] = {127,127,127,16383};         // range per view
int  min_value[] = {0,0,0,0};

uint8_t ccs[8]   = {10,74,71,76,77,93,73,75};   // control change controller codes

#include "meter_image.h"                        // 240x240 PNG bitmap as an array

//--------------------------------------------------------------------------------
// VARIABLES
//--------------------------------------------------------------------------------

PNG       png;                                // PNG decoder 

int16_t   sHandLen;
int16_t  *cached_points[8];

int       curValue [8]  = {0,0,0,0,0,0,0,0};  // value currently on display
int       midiValue[8]  = {0,0,0,0,0,0,0,0};  // value derived from MIDI
int       viewType [8]  = {1,0,0,0,0,0,0,0};  // message type = Control Change

int       nMidi         = 0;
int       midiIn[10];                         // MIDI data buffer  

bool      is_command    = true;               // cheezy way to change displays
int       user_channel  = 1;

uint16_t  png_pixels[240];

//--------------------------------------------------------------------------------
// CODE
//--------------------------------------------------------------------------------

// NAME:        setup
// DESCRIPTION:
void setup(void)
{
    int ch;
  
    Serial.begin(115200);        // for debugging
    Serial1.begin(31250);        // MIDI   
    sHandLen      = 100;
    displays_init();
    for(ch = 0;ch < NUM_DISPLAYS;ch++)
    {
        cached_points[ch] = (int16_t *)malloc((sHandLen + 1) * 2 * 2);
        view_init(ch);
    }
    for(ch = 0;ch < NUM_DISPLAYS;ch++)
    {
        display_select(ch);
        gc9a01_set_on();
    }
    nMidi = 0;     
}

// NAME:        loop
// DESCRIPTION: 
void loop()
{
    int ch;
    int new_value;
 
    midi_service();                         // get new data, update values
    for(ch = 0;ch < NUM_DISPLAYS;ch++)      
    {
         new_value = physics_move(ch,midiValue[ch]); // make needle move realistically  
         if(new_value != curValue[ch])
         {
             curValue[ch] = new_value;
             view_draw(ch);
         }
    }
    user_service();
    delay(1);                               // typical update rate 1 ms
}

//--------------------------------------------------------------------------------
// CODE: USER CONTROL
//--------------------------------------------------------------------------------

// NAME:        user_service
// DESCRIPTION: Service the user serial port, for setting display type 
//              Cheap and cheerful in version 1
//  As new display types are added, need to increase max display type accepted
void user_service(void)
{
    uint8_t c;
    
    if(Serial.available())
    {
        c = Serial.read();
        if(is_command)
        {
            if((c >= '1')
            && (c <= '8'))
            {
                 user_channel = (c - '0');
                 Serial.print("\nDisplay ");
                 Serial.print(user_channel);
                 Serial.print(" selected");                 
                 Serial.print("\nenter display type 0-1");       
                 is_command = false;
            }
            else
                 Serial.print("\nenter 1-8 to select a display");
        }
        else
        {
            if((c >= '0')
            && (c <= '1'))  // this '1' will increase as new display types added
            {
                 viewType[user_channel - 1] = c - '0';
                 Serial.print("\nDisplay type selected");
                 view_init(user_channel - 1);                 
            }
            else
                Serial.print("\ninvalid display type entered");
            is_command                 = true;
        } 
    }
}

//--------------------------------------------------------------------------------
// CODE: MIDI
//--------------------------------------------------------------------------------

// NAME:        midi_service
// DESCRIPTION: Check for, and process, any MIDI bytes received
void midi_service(void)
{
    int ch = 0; 
    int c;
    uint16_t v;
     
    if(Serial1.available())
    {
        c  = Serial1.read();
        if(midi_service_realtime(c))                       // single bytes 
            return;
        if(c & 0x80)                                          // command?
        {
            midiIn[0] = c;
            nMidi     = 1;
        }
        else                                                  // no its data
        if((nMidi == 0)
        && (midiIn[0] & 0x80))
        {
           midiIn[1] = c;                                     // its running status!
           nMidi     = 2;
        }
        else
            midiIn[nMidi++] = c;
        
        switch(midiIn[0] & 0xf0)                               // MIDI command?
        {
            case MIDI_CONTROL:   midi_do_control();    break;                                        
            case MIDI_PITCHBEND: midi_do_pitchbend();  break;               
            default:
               nMidi = 0;
        }   
    }
}

// NAME:        midi_do_control
// DESCRIPTION: Control Change message, capture its value
void midi_do_control(void)
{
    int ch;
    int v;
    
    if(nMidi == 3)
    {
        for(ch = 0;ch < NUM_DISPLAYS;ch++)               // matching controller?
             if((ccs[ch] == midiIn[1])
             && (viewType[ch] == VIEW_CONTROL))
                 break;
        if(ch != NUM_DISPLAYS)
            midiValue[ch] = midiIn[2];           
//        Serial.println(ch);           
        nMidi         = 0;
    }
}

// NAME:        midi_do_pitchbend
// DESCRIPTION: Pitch wheel message, capture value
void midi_do_pitchbend(void)
{
    int ch;
    int v;
        
    if(nMidi == 3)
    {
        v   = midiIn[2];
        v  *= 128;
        v  += midiIn[1];
        for(ch = 0;ch < NUM_DISPLAYS;ch++)    // apply to views
             if(viewType[ch] == VIEW_PITCH)
                 midiValue[ch] = v;
       // Serial.print("\nPitchbend ");
       // Serial.println(midiIn[2]);
        nMidi         = 0;
    }
}
            
// NAME:        midi_service_realtime
// DESCRIPTION: Currently just filters out system realtime
bool midi_service_realtime(byte c)
{
   switch(c)                                             // System Realtime
    {
        case MIDI_STOP:
        case MIDI_START:
        case MIDI_CONTINUE:
        case MIDI_CLOCK:
        case MIDI_ACTIVE:
        case MIDI_RESET:
            nMidi = 0;
            return true;
    }
    return false;
}

//--------------------------------------------------------------------------------
// CODE: PHYSICS
//--------------------------------------------------------------------------------

// physics
// all meters have the same physics behaviour

float curDamping  = 0.96;
float curPull     = 0.002;                    // magnet and spring pull
float velocity[8] = {0,0,0,0,0,0,0,0};        // each meter rate of change
float position[8] = {0,0,0,0,0,0,0,0};        // 0.1

// NAME:        physics_move
// DESCRIPTION:
int physics_move(int ch,int value)           
{
    float vnew;
    float pnew;
    float up;
    float down;
    float accel;
    int  mtype;

   // return value;

    vnew    = velocity[ch];
    up      = value * curPull;           // magnet pull to the right 
    down    = position[ch] * curPull;    // spring pull to the left
    accel   = up - down;
    vnew   += accel;
    vnew   *= curDamping;
    pnew    = position[ch] + vnew;

    mtype = viewType[ch];
    if(pnew < min_value[mtype])            // hit bottom...hard stop
    {
        pnew = min_value[mtype];
        vnew = abs(vnew) * 0.5;
    }
    else
    if(pnew > max_value[mtype])          // pinned...hard stop
    {
       pnew = max_value[mtype];
       vnew = abs(vnew) * -0.5;
    }
    position[ch] = pnew;   
    velocity[ch] = vnew;

    return (int)pnew;
}

//-------------------------------------------------------------------------------
// CODE: VIEW DRAWING
//-------------------------------------------------------------------------------

// NAME:        view_init
// DESCRIPTION:
void view_init(int ch)
{
    display_select(ch);
    midiValue[ch] = curValue[ch] = 0;
    switch(viewType[ch])
    {
        case VIEW_CONTROL:
            png_show();
            draw_needle(ch);
            break;
        case VIEW_PITCH:                     // pitch bend
            midiValue[ch] = curValue[ch] = 8092;
            draw_pitch(ch,false);
            draw_pitch(ch,true);
            break;
        // add more views here
    }
}

// NAME:        view_service
// DESCRIPTION: VU meter draw routine
void view_service(int ch)
{
    if(curValue[ch] != midiValue[ch])    // if moving toward current...
    { 
        curValue[ch] = physics_move(ch,midiValue[ch]);                // make needle move realistically  
        view_draw(ch);
    }
}

// NAME:        view_draw
// DESCRIPTION: Draw an update to the view for channel
void view_draw(int ch)
{    
    display_select(ch);             
    switch(viewType[ch])
    {
        case VIEW_CONTROL: draw_needle (ch);      break;
        case VIEW_PITCH:   draw_pitch  (ch,true); break;
        // add more views here
    }
}

// NAME:        draw_needle
// DESCRIPTION: VU meter draw routine
void draw_needle(int ch)
{
    float angle;
    int16_t   x,y;                         

    angle = 52.5 - 90.0 + ((curValue[ch] * 75.0) / 127.0);   // degrees 
    angle = (angle / 360.0) * 6.2831853;                     // radians
    angle -= RIGHT_ANGLE_RADIAN;
    x  = (cos(angle) * sHandLen) + 120;
    y  = (sin(angle) * sHandLen) + 195;
    draw_and_erase_cached_line(ch,120,195,x,y,BLACK,cached_points[ch],sHandLen + 1,true);
}

// NAME:        draw_pitch
// DESCRIPTION: Pitch bend meter, +- around 12 o'clock
void draw_pitch(int ch,bool isForeground)
{
    float angle;
    int   length;
    int16_t   x,y;                         
  
    if(isForeground)                                 // draw the foreground (needle)
    {
        angle  = curValue[ch] - 8092;
        angle /= 8092;                                    // normalize to -1.0 -> 1.0
        angle *= 130.0;                                   // set the angle CW & CCW
        angle  = (angle / 360.0) * 6.2831853;             // convert to radians
        angle -= RIGHT_ANGLE_RADIAN;
        length = DISPLAY_CENTER - 40;
      
        x    = (cos(angle) * length) + DISPLAY_CENTER;  // calculate the end point
        y    = (sin(angle) * length) + DISPLAY_CENTER;  
        draw_and_erase_cached_line (ch,DISPLAY_CENTER,DISPLAY_CENTER,x,y,BLACK,cached_points[ch],length + 1,false);
  }                                                     // draw it
  else
      gc9a01_fill_screen(WHITE);                        // draw the background 
}

//--------------------------------------------------------------------------------
// CODE: DISPLAYS, HIGH-LEVEL
//--------------------------------------------------------------------------------

void displays_init(void)
{
    int ch;

    gc9a01_init_low();

    pinMode(PIN_S0,OUTPUT);    
    digitalWrite(PIN_S0,0);
    pinMode(PIN_S1,OUTPUT);
    digitalWrite(PIN_S1,0);
    pinMode(PIN_S2,OUTPUT);
    digitalWrite(PIN_S2,0);
    for(ch = 0;ch < NUM_DISPLAYS;ch++)  // initialize all
    {
        display_select(ch);
        delay(2);
        gc9a01_init();
    }
    pinMode(PIN_BL,OUTPUT);
    digitalWrite(PIN_BL,1);            // turn on backlights
    delay(2000);
}

// NAME:        display_select
// DESCRIPTION: Select which display to communicate with via 74HC138
void display_select(int ch)
{
  ch = ch ^ 0x7;                           // invert order left to right
  digitalWrite(PIN_S0,(ch & 0x1) ? 1 : 0);
  digitalWrite(PIN_S1,(ch & 0x2) ? 1 : 0);
  digitalWrite(PIN_S2,(ch & 0x4) ? 1 : 0);
}

//-------------------------------------------------------------------------------
// CODE: PNG BITMAP
//-------------------------------------------------------------------------------

// NAME:        png_show
// DESCRIPTION: Display the PNG
void png_show(void)
{
    int rc;
  
    rc = png.openFLASH ((uint8_t *)img_png_meter_,sizeof(img_png_meter_),png_draw);
    if (rc == PNG_SUCCESS) 
        rc = png.decode(NULL,PNG_FAST_PALETTE);
}

// NAME:        png_draw//PNGDraw
// DESCRIPTION: Draw a line of the PNG
void png_draw(PNGDRAW *pDraw)
{
    png.getLineAsRGB565 (pDraw,png_pixels,PNG_RGB565_BIG_ENDIAN, 0xffffffff); // don't do alpha color blending
    gc9a01_line_show    (0,pDraw->y,pDraw->iWidth,png_pixels);
}

//-------------------------------------------------------------------------------
// CODE: NEEDLE DRAWING
//-------------------------------------------------------------------------------

// NAME:        draw_and_erase_cached_line
// DESCRIPTION:
void draw_and_erase_cached_line(int ch,int16_t x0,int16_t y0,int16_t x1,int16_t y1,int16_t color,int16_t *cache,int16_t cache_len,bool black_clip)
{
    bool    steep;
    int16_t dx,dy;
    int16_t x,y,ox,oy;
    int16_t err;
    int8_t  xstep;
    int8_t  ystep;
    uint16_t i;
   
    steep = _diff(y1,y0) > _diff(x1,x0);
    if(steep)
    {
        _swap_int16_t(x0,y0); 
        _swap_int16_t(x1,y1);
    }
    dx     = _diff(x1,x0);
    dy     = _diff(y1,y0);
    err    = dx / 2;
    xstep  = (x0 < x1) ? 1 : -1;
    ystep  = (y0 < y1) ? 1 : -1;
    x1    += xstep;
    for(i = 0;i <= dx;i++)
    {
        if(steep)
        {
            x = y0;
            y = x0;
        }
        else
        {
            x = x0;
            y = y0;
        }
        ox = *(cache + (i * 2));
        oy = *(cache + (i * 2) + 1);
        
        if(!black_clip
        || (y < BLACK_TOP))
        {     
            gc9a01_draw_pixel(x,y,color);
            if((x != ox)
            || (y != oy))
            {
                if((ox > 0)
                || (oy > 0))
                    gc9a01_draw_pixel(ox,oy,BACKGROUND);
                *(cache + (i * 2))     = x;
                *(cache + (i * 2) + 1) = y;
            }
        }
        
        if(err < dy)
        {
            y0  += ystep;
            err += dx;
        }
        err -= dy;
        x0  += xstep;
    }
  
    for(i = dx +1;i < cache_len;i++)
    {
        ox = *(cache + (i * 2));
        oy = *(cache + (i * 2) + 1);
        if(!black_clip
        || (y < BLACK_TOP))
        {
            if((ox > 0)
            || (oy > 0))
            {
                gc9a01_draw_pixel(ox,oy,BACKGROUND);
                *(cache + (i * 2))     = 0;
                *(cache + (i * 2) + 1) = 0;
            }
        }
    }
}

// NAME:        _diff
// DESCRIPTION: Absolute value of difference
int _diff(int a,int b)
{
  a = a - b;
  if(a < 0)
    a = -a;

  return a;
}
